<?php

declare(strict_types=1);

namespace Drupal\one_time_password\Controller;

use Drupal\user\Controller\UserAuthenticationController as BaseUserAuthenticationController;
use Drupal\user\UserAuthenticationInterface;
use Drupal\user\UserDataInterface;
use Psr\Clock\ClockInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

/**
 * Defines a controller for user authentication.
 */
final class UserAuthenticationController extends BaseUserAuthenticationController {

  const EXCEPTION_MESSAGE = 'Sorry, unrecognized username or password.';

  /**
   * The user data service.
   */
  protected UserDataInterface $userData;

  /**
   * Symfony clock service.
   */
  protected ClockInterface $clock;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->userData = $container->get('user.data');
    $instance->clock = $container->get('one_time_password.clock');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function login(Request $request) {
    $format = $this->getRequestFormat($request);

    $content = $request->getContent();
    $credentials = $this->serializer->decode($content, $format);
    if (!isset($credentials['name'])) {
      throw new BadRequestHttpException(self::EXCEPTION_MESSAGE);
    }
    $account = FALSE;

    if ($this->userAuth instanceof UserAuthenticationInterface) {
      $account = $this->userAuth->lookupAccount($credentials['name']);
    }
    else {
      $accounts = $this->userStorage->loadByProperties(['name' => $credentials['name']]);
      if ($accounts) {
        $account = reset($accounts);
      }
    }
    if ($account && $account->one_time_password->isEmpty()) {
      return parent::login($request);
    }

    if (!$account) {
      throw new BadRequestHttpException(self::EXCEPTION_MESSAGE);
    }

    if (!$request->headers->has('X-OTP')) {
      throw new BadRequestHttpException(self::EXCEPTION_MESSAGE);
    }
    /** @var \Drupal\user\UserInterface $account */
    $uid = (int) $account->id();
    $leeway = 10;
    $last_accepted_window = $this->userData->get('one_time_password', $uid, 'one_time_password_totp_time_window') ?: 1;

    $now = $this->clock->now()->getTimestamp();
    // The lower limit of the current window.
    $min_window = $now - $leeway;

    /** @var \OTPHP\TOTP $one_time_pass */
    $one_time_pass = $account->one_time_password->getOneTimePassword();
    $user_provided_code = $request->headers->get('X-OTP');
    if (
      // No code provided.
      empty($user_provided_code) ||
      // The code used overlapped with a previous window.
      $min_window <= $last_accepted_window ||
      // Invalid code.
      !$one_time_pass->verify($user_provided_code, NULL, $leeway)) {
      throw new BadRequestHttpException(self::EXCEPTION_MESSAGE);
    }
    $response = parent::login($request);
    if ($response->getStatusCode() === Response::HTTP_OK) {
      $this->userData->set('one_time_password', $uid, 'one_time_password_totp_time_window', $now + $one_time_pass->expiresIn());
    }
    return $response;
  }

}
