<?php

declare(strict_types=1);

namespace Drupal\one_time_password\EventSubscriber;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Utility\Crypt;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Routing\UrlGeneratorInterface;
use Drupal\Core\Session\SessionManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\Core\Url;
use Drupal\one_time_password\OtpLoginTrait;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Event subscriber for enforcing TFA with One Time Login links.
 *
 * @internal
 */
final class OneTimeLoginEventSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;
  use OtpLoginTrait;

  /**
   * Constructs a new OneTimeLoginEventSubscriber.
   */
  public function __construct(
    protected RouteMatchInterface $routeMatch,
    protected ConfigFactoryInterface $configFactory,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected TimeInterface $time,
    protected LoggerChannelFactoryInterface $loggerChannelFactory,
    protected PrivateTempStoreFactory $tempStoreFactory,
    protected UrlGeneratorInterface $urlGenerator,
    protected MessengerInterface $messenger,
    TranslationInterface $translation,
    protected SessionManagerInterface $sessionManager,
  ) {
    $this->setStringTranslation($translation);
  }

  /**
   * Redirect One Time Logins when users have TFA for their account.
   */
  public function redirectOneTimeLogin(RequestEvent $event): void {
    if ($this->routeMatch->getRouteName() !== 'user.reset.login') {
      return;
    }

    $request = $event->getRequest();
    $uid = (int) $this->routeMatch->getRawParameter('uid');
    $timestamp = (int) $this->routeMatch->getRawParameter('timestamp');
    $hash = $this->routeMatch->getRawParameter('hash');

    if ($hash === NULL) {
      $this->setRedirectToUserPassPage($event);
      return;
    }

    /** @var \Drupal\user\UserInterface|null $user */
    $user = $this->entityTypeManager->getStorage('user')->load($uid);
    if (
      $user === NULL ||
      !$user->isActive() ||
      !$user->isAuthenticated() ||
      $user->one_time_password->isEmpty()
    ) {
      return;
    }

    // The following code is based on core 9.4.6.
    // @see \Drupal\user\Controller\UserController::resetPassLogin()
    $current = $this->time->getRequestTime();

    // Time out, in seconds, until login URL expires.
    $timeout = $this->configFactory->get('user.settings')->get('password_reset_timeout');
    // No time out for first time login.
    if ($user->getLastLoginTime() && $current - $timestamp > $timeout) {
      $this->messenger->addError($this->t('You have tried to use a one-time login link that has expired. Please request a new one using the form below.'));
      $this->setRedirectToUserPassPage($event);
      return;
    }

    if (($timestamp < $user->getLastLoginTime()) || ($timestamp > $current) || !hash_equals($hash, user_pass_rehash($user, $timestamp))) {
      $this->messenger->addError($this->t('You have tried to use a one-time login link that has either been used or is no longer valid. Please request a new one using the form below.'));
      $this->setRedirectToUserPassPage($event);
      return;
    }

    // CWE-384 Session Fixation protection for EntryForm.
    $this->sessionManager->regenerate();

    $token = Crypt::randomBytesBase64(55);
    $request->getSession()->set('pass_reset_' . $uid, $token);

    // Store UID in order to later verify access to entry form.
    $this->tempStoreFactory->get('one_time_password')->set('otp-entry-uid', $user->id());

    $destination = Url::fromRoute(
      'one_time_password.entry',
      [
        'uid' => $uid,
        'hash' => $this->getLoginHash($user),
      ],
      [
        'query' => ['pass-reset-token' => $token],
      ]
    )
      ->setUrlGenerator($this->urlGenerator)
      ->toString();

    $redirect = new RedirectResponse($destination);
    $event->setResponse($redirect);
  }

  /**
   * Configure the event to redirect to the 'user.pass' page.
   */
  private function setRedirectToUserPassPage(RequestEvent $event): void {
    $destination = Url::fromRoute('user.pass')->setUrlGenerator($this->urlGenerator)->toString();
    $redirect = new RedirectResponse($destination);
    $event->setResponse($redirect);
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    // This event needs to run right after the route matcher.
    $events[KernelEvents::REQUEST][] = ['redirectOneTimeLogin', 31];
    return $events;
  }

}
