<?php

namespace Drupal\one_time_password\Form;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\DependencyInjection\AutowireTrait;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\one_time_password\OtpLoginTrait;
use Drupal\one_time_password\UserLoginEnforce;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;

/**
 * OTP entry form.
 */
final class EntryForm extends FormBase {

  use AutowireTrait;
  use OtpLoginTrait;

  /**
   * Constructs a new EntryForm.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected PrivateTempStoreFactory $tempStore,
  ) {
  }

  const string FORM_ID = 'one_time_password_entry_form';

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return self::FORM_ID;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['one_time_password'] = [
      '#title' => $this->t('One Time Password'),
      '#description' => $this->t('If you have two-factor authentication enabled, enter your one time password.'),
      '#type' => 'textfield',
      '#size' => 6,
      '#maxlength' => 6,
    ];
    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];
    $form['#validate'][] = [static::class, 'validateOneTimePassword'];
    return $form;
  }

  /**
   * Validate callback to verify the one time password code.
   */
  public static function validateOneTimePassword(array &$form, FormStateInterface $form_state): void {
    // A user ID means the user was able to authenticate. We only care about
    // validating the TFA code if the user was already authenticated by a
    // password. If the user has not authenticated by a password, the form
    // validating will stop login anyway.
    $uid = \Drupal::routeMatch()->getParameter('uid');
    if (!$uid) {
      return;
    }

    // If the one_time_password field is empty, don't validate the users one
    // time password code. They haven't opted into using one time passwords.
    $user = User::load($uid);
    if ($user->one_time_password->isEmpty()) {
      return;
    }

    /** @var \Drupal\Core\Flood\FloodInterface $flood */
    $flood = \Drupal::service('flood');
    $ip_address = \Drupal::requestStack()->getCurrentRequest()->getClientIp();

    // Check if the user has attempted too many OTP guesses from a single IP
    // address or for a specific user.
    $flood_is_allowed_uid = $flood->isAllowed(UserLoginEnforce::FLOOD_NAME_UID, UserLoginEnforce::FLOOD_THRESHOLD, UserLoginEnforce::FLOOD_WINDOW, $uid);
    $flood_is_allowed_ip = $flood->isAllowed(UserLoginEnforce::FLOOD_NAME_IP, UserLoginEnforce::FLOOD_THRESHOLD, UserLoginEnforce::FLOOD_WINDOW, $ip_address);
    if (!$flood_is_allowed_uid || !$flood_is_allowed_ip) {
      $form_state->setErrorByName('one_time_password', \t('There have been too many incorrect one time passwords entered.'));
      // Don't attempt to validate the OTP if the flood control has kicked in.
      return;
    }

    // Get the one time password object from the user and verify it against the
    // one time password that was submitted. The flood table validation runs
    // before this validation, so this is not a brute-force vector. We only see
    // this message if the user has successfully validated their password and
    // have not tripped a flood threshold.
    /** @var \OTPHP\TOTP $one_time_pass */
    $one_time_pass = $user->one_time_password->getOneTimePassword();
    $user_provided_code = $form_state->getValue('one_time_password');
    /** @var \Drupal\user\UserDataInterface $user_data */
    $user_data = \Drupal::service('user.data');
    // The upper limit of the last TOTP window.
    $last_accepted_window = $user_data->get('one_time_password', $uid, 'one_time_password_totp_time_window') ?: 1;
    $leeway = 10;
    $now = \Drupal::service('one_time_password.clock')->now()->getTimestamp();
    // The lower limit of the current window.
    $min_window = $now - $leeway;
    if (
      // No code provided.
      empty($user_provided_code) ||
      // The code used overlapped with a previous window.
      $min_window <= $last_accepted_window ||
      // Invalid code.
      !$one_time_pass->verify($user_provided_code, NULL, $leeway)
    ) {
      $form_state->setErrorByName('one_time_password', \t('The entered two-factor authentication code is incorrect.'));

      // If the OTP has been entered incorrectly, add an entry to the flood
      // table. Note: this is NOT registered if the user has already exceeded
      // the flood thresholds. This means logged repeated attempts will not
      // ever exceed the flood threshold.
      $flood->register(UserLoginEnforce::FLOOD_NAME_UID, UserLoginEnforce::FLOOD_WINDOW, $uid);
      $flood->register(UserLoginEnforce::FLOOD_NAME_IP, UserLoginEnforce::FLOOD_WINDOW, $ip_address);
    }
    else {
      // Store the upper limit of the current window.
      $user_data->set('one_time_password', $uid, 'one_time_password_totp_time_window', $now + $one_time_pass->expiresIn());
    }
  }

  /**
   * If the form is passes validation, the user should get logged in.
   *
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $uid = $this->getRouteMatch()->getParameter('uid');
    if (!$uid) {
      return;
    }
    /** @var \Drupal\user\UserInterface $account */
    $account = $this->entityTypeManager->getStorage('user')->load($uid);

    // A destination was set, probably on an exception controller.
    if (!$this->getRequest()->request->has('destination')) {
      $form_state->setRedirect(
        'entity.user.canonical',
        ['user' => $account->id()]
      );
    }
    else {
      $this->getRequest()->query->set('destination', $this->getRequest()->request->get('destination'));
    }
    user_login_finalize($account);
  }

  /**
   * Check access to the form.
   */
  public function checkAccess(RouteMatchInterface $route, AccountInterface $account): AccessResultInterface {
    $access = AccessResult::allowed();
    // Use uids here instead of user objects to prevent enumeration attacks.
    $uid = (int) $route->getParameter('uid');

    // If stored uid is invalid, the tfa process didn't start in this session.
    $temp_store = $this->tempStore->get('one_time_password');
    $uid_check = $temp_store->get('otp-entry-uid');
    if (
      !$uid_check ||
      !is_numeric($uid_check) ||
      ($uid !== (int) $uid_check)
    ) {
      return $access->andIf(AccessResult::forbidden('Invalid session.'));
    }

    // Attempt to retrieve a user from the uid.
    $user = User::load($uid);
    if (!$user instanceof UserInterface) {
      return $access->andIf(AccessResult::forbidden('Invalid user.'));
    }

    if ($user->one_time_password->isEmpty()) {
      return $access->andIf(AccessResult::forbidden('User does not have TFA enabled.'));
    }

    // If the login hash doesn't match, forbid access.
    if ($this->getLoginHash($user) !== $route->getParameter('hash')) {
      return $access->andIf(AccessResult::forbidden('Invalid hash value.'));
    }

    return $access;
  }

}
