<?php

declare(strict_types=1);

namespace Drupal\Tests\one_time_password\Functional;

use Drupal\Core\Url;
use Drupal\one_time_password\OtpLoginTrait;
use Drupal\one_time_password\Plugin\Field\FieldType\ProvisioningUriItemList;
use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\user\Traits\UserCreationTrait;
use Drupal\user\UserInterface;
use OTPHP\TOTP;

/**
 * Test ULIs.
 *
 * @coversDefaultClass \Drupal\one_time_password\Form\EntryForm
 * @group one_time_password
 */
class EntryFormTest extends BrowserTestBase {

  use UserCreationTrait;
  use OtpLoginTrait;

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'one_time_password',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    $this->setPrivateKey(\Drupal::service('private_key'));
  }

  /**
   * Test login via ULI.
   */
  public function testUliLogin(): void {
    // Create a user with TFA enabled.
    $test_user_with_tfa = $this->createUser();
    \assert($test_user_with_tfa instanceof UserInterface);
    \assert($test_user_with_tfa->one_time_password instanceof ProvisioningUriItemList);
    $test_user_with_tfa->one_time_password->regenerateOneTimePassword();
    $otp_object = $test_user_with_tfa->one_time_password->getOneTimePassword();
    \assert($otp_object instanceof TOTP);
    $test_user_with_tfa->save();

    $test_user_with_tfa2 = $this->createUser();
    \assert($test_user_with_tfa2 instanceof UserInterface);
    \assert($test_user_with_tfa2->one_time_password instanceof ProvisioningUriItemList);
    $test_user_with_tfa2->one_time_password->regenerateOneTimePassword();
    $test_user_with_tfa2->save();

    $test_user_no_tfa = $this->createUser();
    \assert($test_user_no_tfa instanceof UserInterface);
    \assert($test_user_no_tfa->one_time_password instanceof ProvisioningUriItemList);
    $test_user_no_tfa->save();

    // Generate a ULI for the user.
    $uri = user_pass_reset_url($test_user_with_tfa);
    $this->drupalGet($uri);
    $this->submitForm([], 'Log in');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->elementExists('css', 'h1:contains("Two-Factor Authentication")');
    // Login with the ULI.
    $this->submitForm([
      'one_time_password' => $otp_object->now(),
    ], 'Submit');
    $test_user_with_tfa->sessionId = $this->getSession()->getCookie($this->getSessionName());
    $this->assertTrue($this->drupalUserIsLoggedIn($test_user_with_tfa));

    // Logout, try to manipulate the URI and check access is denied.
    $this->drupalLogout();

    // Use random hash with a valid uid.
    $this->drupalGet(Url::fromRoute('one_time_password.entry', [
      'uid' => $test_user_with_tfa->id(),
      'hash' => $this->randomMachineName(16),
    ])->toString());
    $this->assertSession()->statusCodeEquals(403);

    // Use a random uid with a valid hash.
    $this->drupalGet(Url::fromRoute('one_time_password.entry', [
      'uid' => 7,
      'hash' => $this->getLoginHash($test_user_with_tfa),
    ])->toString());
    $this->assertSession()->statusCodeEquals(403);

    // Valid user id and hash, but user has not configured tfa.
    $this->drupalGet(Url::fromRoute('one_time_password.entry', [
      'uid' => $test_user_no_tfa->id(),
      'hash' => $this->getLoginHash($test_user_no_tfa),
    ])->toString());
    $this->assertSession()->statusCodeEquals(403);

    // Check ULI for a user with no TFA.
    $uri = user_pass_reset_url($test_user_no_tfa);
    $this->drupalGet($uri);
    $this->assertSession()->pageTextContains('This is a one-time login for');
    $this->submitForm([], 'Log in');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->elementExists('css', \sprintf('h1:contains("%s")', $test_user_no_tfa->getDisplayName()));
  }

}
