<?php

declare(strict_types=1);

namespace Drupal\Tests\one_time_password\Functional;

use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\one_time_password\Controller\UserAuthenticationController;
use Drupal\one_time_password\Plugin\Field\FieldType\ProvisioningUriItemList;
use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\user\Traits\UserCreationTrait;
use Drupal\user\UserInterface;
use GuzzleHttp\Cookie\CookieJar;
use OTPHP\TOTP;
use Psr\Http\Message\ResponseInterface;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Symfony\Component\Serializer\Encoder\XmlEncoder;
use Symfony\Component\Serializer\Serializer;

/**
 * Test the actual process of logging in with TFA.
 *
 * @coversDefaultClass \Drupal\one_time_password\UserLoginEnforce
 * @group one_time_password
 */
class UserLoginEnforceTest extends BrowserTestBase {

  use UserCreationTrait;

  /**
   * The cookie jar.
   *
   * @var \GuzzleHttp\Cookie\CookieJar
   */
  protected $cookies;

  /**
   * The serializer.
   *
   * @var \Symfony\Component\Serializer\Serializer
   */
  protected $serializer;

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'one_time_password',
  ];

  /**
   * Test the user login with TFA codes.
   */
  public function testUserLogin(): void {
    // Users without TFA enabled should be able to login as normal.
    $test_user = $this->createUser();
    \assert($test_user instanceof UserInterface);
    $this->drupalLogin($test_user);
    $this->drupalLogout();

    // Create a user with TFA enabled.
    $test_user_with_tfa = $this->createUser();
    \assert($test_user_with_tfa instanceof UserInterface);
    \assert($test_user_with_tfa->one_time_password instanceof ProvisioningUriItemList);
    $test_user_with_tfa->one_time_password->regenerateOneTimePassword();
    $otp_object = $test_user_with_tfa->one_time_password->getOneTimePassword();
    \assert($otp_object instanceof TOTP);
    $test_user_with_tfa->save();

    // Attempt to login with a user who has TFA enabled.
    $this->drupalGet('user/login');
    $this->submitForm([
      'name' => $test_user_with_tfa->getAccountName(),
      'pass' => $test_user_with_tfa->passRaw,
    ], 'Log in');
    $this->assertSession()->pageTextContains('The entered two-factor authentication code is incorrect.');

    // Attempt to login with an incorrect TFA code.
    $this->drupalGet('user/login');
    $this->submitForm([
      'name' => $test_user_with_tfa->getAccountName(),
      'pass' => $test_user_with_tfa->passRaw,
      'one_time_password' => '123456',
    ], 'Log in');
    $this->assertSession()->pageTextContains('The entered two-factor authentication code is incorrect.');

    // Attempt to login with the correct TFA code and ensure the user is
    // correctly logged in.
    $this->drupalGet('user/login');
    $otp_code = $otp_object->now();
    $this->submitForm([
      'name' => $test_user_with_tfa->getAccountName(),
      'pass' => $test_user_with_tfa->passRaw,
      'one_time_password' => $otp_code,
    ], 'Log in');
    $test_user_with_tfa->sessionId = $this->getSession()->getCookie($this->getSessionName());
    $this->assertTrue($this->drupalUserIsLoggedIn($test_user_with_tfa), 'User was logged in successfully with a TFA code.');

    // Try to reuse the same login code after a 21 second sleep.
    sleep(21);
    $this->drupalLogout();
    $this->drupalGet('user/login');
    $this->submitForm([
      'name' => $test_user_with_tfa->getAccountName(),
      'pass' => $test_user_with_tfa->passRaw,
      'one_time_password' => $otp_code,
    ], 'Log in');
    $test_user_with_tfa->sessionId = $this->getSession()->getCookie($this->getSessionName());
    $this->assertSession()->pageTextContains('The entered two-factor authentication code is incorrect.');
    $this->assertFalse($this->drupalUserIsLoggedIn($test_user_with_tfa));
  }

  /**
   * Log the user in through the 'user.login.http' route.
   */
  protected function doLogin(AccountInterface $account, ?string $one_time_password = NULL): ResponseInterface {
    $user_login_url = Url::fromRoute('user.login.http')
      ->setRouteParameter('_format', 'json')
      ->setAbsolute();
    $result = \Drupal::httpClient()->post($user_login_url->toString(), [
      'body' => $this->serializer->encode([
        'name' => $account->getAccountName(),
        'pass' => $account->passRaw,
      ], 'json'),
      'headers' => \array_filter([
        'Accept' => "application/json",
        'X-OTP' => $one_time_password,
      ]),
      'http_errors' => FALSE,
      'cookies' => $this->cookies,
    ]);
    return $result;
  }

  /**
   * Assert the HTTP response code and message.
   *
   * @phpstan-param array<string> $message_keys
   */
  protected function assertHttpResponseWithMessage(ResponseInterface $response, int $expected_code, string $expected_message, array $message_keys = ['message']): void {
    self::assertEquals($expected_code, $response->getStatusCode());
    $body = Json::decode((string) $response->getBody());
    self::assertEquals($expected_message, NestedArray::getValue($body, $message_keys));
  }

  /**
   * Test the user login with TFA codes via HTTP.
   */
  public function testUserLoginHttp(): void {
    $this->cookies = new CookieJar();
    $encoders = [new JsonEncoder(), new XmlEncoder()];
    $this->serializer = new Serializer([], $encoders);

    // Users without TFA enabled should be able to login as normal.
    $test_user = $this->createUser();
    $this->assertHttpResponseWithMessage(
      $this->doLogin($test_user),
      Response::HTTP_OK,
      $test_user->getAccountName(),
      ['current_user', 'name'],
    );

    // Logout.
    $this->cookies = new CookieJar();

    // Create a user with TFA enabled.
    $test_user_with_tfa = $this->createUser();
    \assert($test_user_with_tfa instanceof UserInterface);
    $test_user_with_tfa->one_time_password->regenerateOneTimePassword();
    $otp_object = $test_user_with_tfa->one_time_password->getOneTimePassword();
    $test_user_with_tfa->save();

    // Attempt to login with a user who has TFA enabled.
    $this->assertHttpResponseWithMessage($this->doLogin($test_user_with_tfa), Response::HTTP_BAD_REQUEST, UserAuthenticationController::EXCEPTION_MESSAGE);

    // Attempt to login with an incorrect TFA code.
    $this->assertHttpResponseWithMessage($this->doLogin($test_user_with_tfa, '123456'), Response::HTTP_BAD_REQUEST, UserAuthenticationController::EXCEPTION_MESSAGE);

    // Attempt to login with the correct TFA code and ensure the user is
    // correctly logged in.
    $code = $otp_object->now();
    $this->assertHttpResponseWithMessage(
      $this->doLogin($test_user_with_tfa, $code),
      Response::HTTP_OK,
      $test_user_with_tfa->getAccountName(),
      ['current_user', 'name']
    );

    $this->cookies = new CookieJar();
    sleep(21);

    // Attempt to reuse the same code.
    $this->assertHttpResponseWithMessage($this->doLogin($test_user_with_tfa, $code), Response::HTTP_BAD_REQUEST, UserAuthenticationController::EXCEPTION_MESSAGE);
  }

  /**
   * Test the user login with invalid credentials.
   */
  public function testUserLoginHttpWithInvalidName(): void {
    $this->cookies = new CookieJar();
    $encoders = [new JsonEncoder(), new XmlEncoder()];
    $this->serializer = new Serializer([], $encoders);
    // This is mostly copied from doLogin, but we don't have an account
    // object here, so duplicating the code here rather than adjusting
    // the method signature to be more open.
    $user_login_url = Url::fromRoute('user.login.http')
      ->setRouteParameter('_format', 'json')
      ->setAbsolute();
    $result = \Drupal::httpClient()->post($user_login_url->toString(), [
      'body' => $this->serializer->encode([
        'name' => $this->randomMachineName(),
        'pass' => $this->randomMachineName(),
      ], 'json'),
      'headers' => \array_filter([
        'Accept' => "application/json",
        'X-OTP' => $this->randomMachineName(),
      ]),
      'http_errors' => FALSE,
      'cookies' => $this->cookies,
    ]);
    $this->assertHttpResponseWithMessage($result, Response::HTTP_BAD_REQUEST, UserAuthenticationController::EXCEPTION_MESSAGE);
  }

  /**
   * Test the wider login window works.
   */
  public function testLoginWindow(): void {
    $test_user_with_tfa = $this->createUser();
    \assert($test_user_with_tfa instanceof UserInterface);
    \assert($test_user_with_tfa->one_time_password instanceof ProvisioningUriItemList);
    $test_user_with_tfa->one_time_password->regenerateOneTimePassword();
    /** @var \OTPHP\TOTP $otp_object */
    $otp_object = $test_user_with_tfa->one_time_password->getOneTimePassword();
    $test_user_with_tfa->save();

    // Attempt to login with the code that would have been valid about three
    // minutes ago.
    $this->drupalGet('user/login');
    /** @var \Symfony\Component\Clock\Clock $clock */
    $clock = \Drupal::service('one_time_password.clock');
    $this->submitForm([
      'name' => $test_user_with_tfa->getAccountName(),
      'pass' => $test_user_with_tfa->passRaw,
      'one_time_password' => $otp_object->at($clock->now()->modify('-3 minutes')->getTimestamp()),
    ], 'Log in');

    $test_user_with_tfa->sessionId = $this->getSession()->getCookie($this->getSessionName());
    $this->assertFalse($this->drupalUserIsLoggedIn($test_user_with_tfa), 'User was not logged in successfully with a TFA code that was valid three minutes ago.');
  }

  /**
   * Test no information disclosure through the flood table.
   */
  public function testFloodTableInformationDisclosure(): void {
    $this->config('user.flood')
      ->set('user_limit', 3)
      ->save();

    $tfa_user = $this->drupalCreateUser();
    \assert($tfa_user instanceof UserInterface);
    \assert($tfa_user->one_time_password instanceof ProvisioningUriItemList);
    $tfa_user->one_time_password->regenerateOneTimePassword();
    $tfa_user->save();

    // Login 6 times incorrectly to trigger a flood warning from core as well
    // as the OTP threshold.
    // phpcs:ignore
    foreach (\range(0, 5) as $i) {
      $this->drupalGet('user/login');
      $this->submitForm([
        'name' => $tfa_user->getAccountName(),
        'pass' => 'incorrect password',
      ], 'Log in');
    }

    // Login with the correct password, which flood will deny, but ensure the
    // incorrect TFA code doesn't give us away.
    $this->drupalGet('user/login');
    $this->submitForm([
      'name' => $tfa_user->getAccountName(),
      'pass' => $tfa_user->passRaw,
      'one_time_password' => '123',
    ], 'Log in');
    $this->assertSession()->pageTextContains('There have been more than 3 failed login attempts');
    // Messages from OTP should not indicate if a password was guessed correctly
    // or not. Verify none of the messages we send are visible.
    $this->assertSession()->pageTextNotContains('The entered two-factor authentication code is incorrect.');
    $this->assertSession()->pageTextNotContains('There have been too many incorrect one time passwords entered.');
  }

  /**
   * Test the flood control.
   */
  public function testOtpFloodControl(): void {
    $tfa_user = $this->drupalCreateUser();
    \assert($tfa_user instanceof UserInterface);
    \assert($tfa_user->one_time_password instanceof ProvisioningUriItemList);
    $tfa_user->one_time_password->regenerateOneTimePassword();
    $tfa_user->save();

    $second_tfa_user = $this->drupalCreateUser();
    \assert($second_tfa_user instanceof UserInterface);
    \assert($second_tfa_user->one_time_password instanceof ProvisioningUriItemList);
    $second_tfa_user->one_time_password->regenerateOneTimePassword();
    $second_tfa_user->save();

    // Login enough times with the correct password to trigger a flood control
    // warning from OTP.
    // phpcs:ignore
    foreach (\range(0, 5) as $i) {
      $this->drupalGet('user/login');
      $this->submitForm([
        'name' => $tfa_user->getAccountName(),
        'pass' => $tfa_user->passRaw,
        'one_time_password' => '123',
      ], 'Log in');
    }
    $this->assertSession()->pageTextContains('There have been too many incorrect one time passwords entered.');

    // Login with a second user to verify there is a per ip address limit.
    $this->drupalGet('user/login');
    $this->submitForm([
      'name' => $second_tfa_user->getAccountName(),
      'pass' => $second_tfa_user->passRaw,
      'one_time_password' => '123',
    ], 'Log in');
    $this->assertSession()->pageTextContains('There have been too many incorrect one time passwords entered.');

    // Update all flood ip address identifiers and attempt to login with the
    // first user to verify there is a per-user limit.
    \Drupal::database()
      ->update('flood')
      ->condition('event', 'one_time_password.ip')
      ->fields(['identifier' => 'not a real ip'])
      ->execute();
    $this->drupalGet('user/login');
    $this->submitForm([
      'name' => $tfa_user->getAccountName(),
      'pass' => $tfa_user->passRaw,
      'one_time_password' => '123',
    ], 'Log in');
    $this->assertSession()->pageTextContains('There have been too many incorrect one time passwords entered.');
  }

  /**
   * Test the flood control with repeatedly correct credentials.
   */
  public function testFloodControlWithCorrectCredentials(): void {
    $tfa_user = $this->drupalCreateUser();

    \assert($tfa_user instanceof UserInterface);
    \assert($tfa_user->one_time_password instanceof ProvisioningUriItemList);
    $tfa_user->one_time_password->regenerateOneTimePassword();
    $tfa_user->save();

    $otp_object = $tfa_user->one_time_password->getOneTimePassword();
    \assert($otp_object instanceof TOTP);

    // Login and logout enough times to trigger the flood threshold.
    // phpcs:ignore
    $user_data = \Drupal::service('user.data');
    $uid = (int) $tfa_user->id();
    foreach (\range(0, 6) as $i) {
      $this->drupalGet('user/login');
      $this->submitForm([
        'name' => $tfa_user->getAccountName(),
        'pass' => $tfa_user->passRaw,
        'one_time_password' => $otp_object->now(),
      ], 'Log in');
      $tfa_user->sessionId = $this->getSession()->getCookie($this->getSessionName());
      $this->assertTrue($this->drupalUserIsLoggedIn($tfa_user), 'User was logged in successfully with a TFA code.');
      $this->drupalLogout();

      // Bypass replay protection so we can keep submitting the same OTP::now
      // code without having to sleep for multiple minutes.
      $user_data->delete('one_time_password', $uid, 'one_time_password_totp_time_window');
    }
  }

}
